"""Authentication helper functions
"""
import os
import re
from flask import request
from .jwt import jwt_manager
from flask_jwt_extended import jwt_required, get_jwt_identity


# Accept user name passed in Basic Auth header
# (password has to checkded before!)
ALLOW_BASIC_AUTH_USER = os.environ.get('ALLOW_BASIC_AUTH_USER', 'False') \
    .lower() in ('t', 'true')


def auth_manager(app, api=None):
    """Authentication setup for Flask app"""
    # Setup the Flask-JWT-Extended extension
    return jwt_manager(app, api)


def optional_auth(fn):
    """Authentication view decorator"""
    return jwt_required(optional=True)(fn)


def get_identity():
    """Get identity (username or dict with username and groups)
       or optional pre-authenticated basic auth user"""
    identity = get_jwt_identity()
    if not identity and ALLOW_BASIC_AUTH_USER:
        auth = request.authorization
        if auth:
            # We don't check password, already authenticated!
            identity = auth.username
    return identity


def get_username(identity):
    """Extract username from identity"""
    username = None
    if identity:
        if isinstance(identity, dict):
            username = identity.get('username')
        else:
            # identity is username
            username = identity
    return username


def get_groups(identity):
    """Extract user groups from identity"""
    groups = []
    if identity:
        if isinstance(identity, dict):
            groups = identity.get('groups', [])
            group = identity.get('group')
            if group:
                groups.append(group)
    return groups


class GroupNameMapper:
    """Group name mapping with regular expressions"""

    def __init__(self, default=''):
        group_mappings = os.environ.get('GROUP_MAPPINGS', default)

        def collect(mapping):
            regex, *replacement = mapping.split('~', 1)
            return (re.compile(regex), replacement[0])

        self.group_mappings = list(
            map(collect, group_mappings.split('#'))) if group_mappings else []

    def mapped_group(self, group):
        # LDAP servers my return a group as list object
        if isinstance(group, list):
            group = ' '.join(group)
        for (regex, replacement) in self.group_mappings:
            if regex.match(group):
                return regex.sub(replacement, group)
        return group

# Usage examples:
# mapper = GroupNameMapper('ship_crew~crew#gis.role.(.*)~\\1')
# print(mapper.mapped_group('ship_crew'))
# print(mapper.mapped_group('gis.role.admin'))
# print(mapper.mapped_group('gis.role.user.group1'))
# print(mapper.mapped_group('gis.none'))



from flask_login import UserMixin
from sqlalchemy import MetaData
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import Session, backref, relationship
from werkzeug.security import generate_password_hash, check_password_hash


class ConfigModels():
    """ConfigModels class

    Provide SQLAlchemy ORM models for ConfigDB queries.
    """

    def __init__(self, db_engine, conn_str=None, extra_tables=[]):
        """Constructor

        :param DatabaseEngine db_engine: Database engine with DB connections
        """
        if conn_str:
            self.engine = db_engine.db_engine(conn_str)
        else:
            self.engine = db_engine.config_db()

        # init models
        self.base = None
        self.custom_models = {}
        self.init_models(extra_tables)

    def session(self):
        """Create a new session."""
        return Session(self.engine)

    def model(self, name):
        """Get SQLAlchemy model.

        :param str name: Table name of model
        """
        # get automap model or custom model
        if name == 'users':
            return self.user_model
        else:
            return self.base.classes.get(name) or self.custom_models.get(name)

    def init_models(self, extra_tables):
        """Setup SQLAlchemy ORM models."""

        # Generate required models from ConfigDB using automap
        TABLES = [
            'users', 'user_infos', 'groups', 'roles',
            'groups_users', 'users_roles', 'groups_roles',
            'resources', 'resource_types', 'permissions',
            'registrable_groups', 'registration_requests',
            'last_update'
        ] + extra_tables

        def table_selector(table_name, meta_data):
            return (table_name in TABLES)

        metadata = MetaData()
        metadata.reflect(self.engine, schema='qwc_config', only=table_selector)
        Base = automap_base(metadata=metadata)

        # setup user model for flask_login
        class User(UserMixin, Base):
            __tablename__ = 'users'
            __table_args__ = ({"schema": "qwc_config"})

            def set_password(self, password):
                self.password_hash = generate_password_hash(password, method='pbkdf2')

            def check_password(self, password):
                return check_password_hash(self.password_hash, password)

        Base.prepare()

        self.base = Base
        self.user_model = User

        UserInfo = Base.classes.user_infos
        Group = Base.classes.groups
        Role = Base.classes.roles

        # single user info per user
        User.user_info = relationship(
            UserInfo,
            uselist=False,
            back_populates="user",
            overlaps="user_infos_collection"
        )

        # sorted user groups
        User.sorted_groups = relationship(
            Group,
            secondary='qwc_config.groups_users',
            order_by=Group.name,
            # avoid duplicate group DELETE on User delete
            viewonly=True
        )
        # sorted user roles
        User.sorted_roles = relationship(
            Role,
            secondary='qwc_config.users_roles',
            order_by=Role.name,
            # avoid duplicate role DELETE on User delete
            viewonly=True
        )

        Group.users_collection = relationship(
            User,
            secondary='qwc_config.groups_users',
            overlaps="groups_collection,user_collection"
        )

        # sorted group users
        Group.sorted_users = relationship(
            User,
            secondary='qwc_config.groups_users',
            order_by=User.name,
            # avoid duplicate user DELETE on Group delete
            viewonly=True
        )
        # sorted group roles
        Group.sorted_roles = relationship(
            Role,
            secondary='qwc_config.groups_roles',
            order_by=Role.name,
            # avoid duplicate role DELETE on Group delete
            viewonly=True
        )

        Role.users_collection = relationship(
            User,
            secondary='qwc_config.users_roles',
            overlaps="roles_collection,user_collection"
        )

        # sorted role users
        Role.sorted_users = relationship(
            User,
            secondary='qwc_config.users_roles',
            order_by=User.name,
            # avoid duplicate user DELETE on Role delete
            viewonly=True
        )
        # sorted roles groups
        Role.sorted_groups = relationship(
            Group,
            secondary='qwc_config.groups_roles',
            order_by=Group.name,
            # avoid duplicate group DELETE on Role delete
            viewonly=True
        )

        # sorted resource children and parent
        Resource = Base.classes.resources
        Resource.children = relationship(
            Resource,
            foreign_keys=[Resource.parent_id],
            order_by=Resource.name,
            backref=backref(
                'parent', remote_side=[Resource.id],
                overlaps="resources,resources_collection"
            ),
            overlaps="resources,resources_collection"
        )

        # resource type
        ResourceType = Base.classes.resource_types
        Resource.resource_type = relationship(
            ResourceType,
            overlaps="resource_types,resources_collection"
        )

        # permission role and resource with singular attribute names
        Permission = Base.classes.permissions
        Permission.role = relationship(
            Role,
            overlaps="permissions_collection,roles"
        )
        Permission.resource = relationship(
            Resource,
            overlaps="permissions_collection,resources"
        )

        # group for registrable group with singular attribute name
        RegistrableGroup = Base.classes.registrable_groups
        RegistrableGroup.group = relationship(
            Group,
            overlaps="groups,registrable_groups_collection"
        )

        # user and registrable group for registration request with
        # singular attribute names
        RegistrationRequest = Base.classes.registration_requests
        RegistrationRequest.user = relationship(
            User,
            overlaps="registration_requests"
        )
        RegistrationRequest.registrable_group = relationship(
            RegistrableGroup,
            overlaps="registrable_groups,registration_requests_collection"
        )

        # user registration requests
        User.registration_requests = relationship(
            RegistrationRequest,
            # destroy registration requests on user delete
            cascade="save-update, merge, delete"
        )



import glob
import json
import os


class Translator:
    """Class for translating strings via json files"""

    def __init__(self, app, request):
        """Constructor.

        :param object app: The Flask app
        :param object request: The Flask requst
        """

        supported_locales = list(map(
            lambda path: os.path.basename(path)[0:-5],
            glob.glob('translations/*.json')
        ))

        DEFAULT_LOCALE = os.environ.get('DEFAULT_LOCALE', 'en')
        locale = request.accept_languages.best_match(supported_locales) or DEFAULT_LOCALE

        self.translations = {}
        try:
            path = os.path.join(app.root_path, 'translations/%s.json' % locale)
            with open(path, 'r') as f:
                self.translations = json.load(f)
        except Exception as e:
            app.logger.error(
                "Failed to load translation strings for locale '%s' from %s, loading default locale\n%s"
                % (locale, path, e)
            )
            path = os.path.join(app.root_path, 'translations/%s.json' % DEFAULT_LOCALE)
            with open(path, 'r') as f:
                self.translations = json.load(f)

    def tr(self, msgId):
        """Translate a string.

        :param str msgId: The message id
        """

        parts = msgId.split('.')
        lookup = self.translations
        for part in parts:
            if isinstance(lookup, dict):
                # get next lookup level
                lookup = lookup.get(part)
            else:
                # lookup level too deep
                lookup = None
            if lookup is None:
                # return input msgId if not found
                lookup = msgId
                break

        return lookup



import os
import datetime
import json
from flask_jwt_extended import JWTManager, unset_jwt_cookies
from flask import redirect, request
from jwt.exceptions import PyJWTError


def jwt_manager(app, api=None):
    """Setup Flask-JWT-Extended extension for services
       with authenticated access"""
    # https://flask-jwt-extended.readthedocs.io/en/stable/options
    app.config['JWT_TOKEN_LOCATION'] = ['headers', 'cookies']
    app.config['JWT_ACCESS_COOKIE_NAME'] = os.environ.get(
        'JWT_ACCESS_COOKIE_NAME', 'access_token_cookie')
    app.config['JWT_COOKIE_CSRF_PROTECT'] = str(os.environ.get(
        'JWT_COOKIE_CSRF_PROTECT', 'False')).upper() == "TRUE"
    app.config['JWT_CSRF_CHECK_FORM'] = True
    app.config['JWT_SECRET_KEY'] = os.environ.get(
        'JWT_SECRET_KEY', os.urandom(24))
    app.config['JWT_ACCESS_COOKIE_PATH'] = os.environ.get(
        'JWT_ACCESS_COOKIE_PATH', '/')

    jwt = JWTManager(app)

    @app.after_request
    def handle_jwt_exceptions(resp):
        # If error is a JWT error, unset JWT cookies and redirect to requested URL
        if resp.status_code == 500 and resp.content_type == "application/json":
            data = json.loads(resp.data)
            if "message" in data and data["message"].startswith("jwtexception:"):
                resp = redirect(request.url)
                unset_jwt_cookies(resp)
                return resp

        return resp

    @app.before_request
    def set_jwt_access_cookie_path():
        # If app.session_interface is a TenantSessionInterface, JWT_ACCESS_COOKIE_PATH is set regarding tenant
        # Else it takes value from SESSION_COOKIE_PATH config var, or APPLICATION_ROOT or /
        # https://flask.palletsprojects.com/en/3.0.x/api/#flask.sessions.SessionInterface.get_cookie_path
        from .tenant_handler import TenantSessionInterface
        if isinstance(app.session_interface, TenantSessionInterface):
            if app.session_interface.is_multi():
                app.session_interface.get_cookie_path(app)

    if api:

        @api.errorhandler
        def restplus_error_handler(error):
            # If error is a JWT error, prefix message with jwtexception: so that it
            # can be recognized in handle_jwt_exceptions above
            if isinstance(error, PyJWTError):
                return {"message": "jwtexception:%s" % str(error)}

            return {}

    @jwt.expired_token_loader
    def handle_expired_token(jwtheader, jwtdata):
        # Unset cookies and redirect to requested page on expired token
        app.logger.warn("Expired token")
        resp = redirect(request.url)
        unset_jwt_cookies(resp)
        return resp

    @jwt.invalid_token_loader
    def handle_invalid_token(err):
        # Unset cookies and redirect to requested page on token error
        app.logger.warn("Invalid token: %s" % str(err))
        resp = redirect(request.url)
        unset_jwt_cookies(resp)
        return resp

    @jwt.unauthorized_loader
    def unauthorized(err):
        # Redirect to requested page on authorized error (i.e. CSRF token error)
        app.logger.warn("Unauthorized: %s" % str(err))
        return redirect(request.url)

    return jwt



import time
import copy


class ExpiringDict:
    """Dict for values where each key will expire after some time."""

    def __init__(self):
        """Constructor"""
        self.cache = {}

    def set(self, key, value, duration=300):
        """Store value under key until expiry.

        :param str key: Key for value
        :param obj value: Value to store
        :param int duration: Time in seconds until expiry (default: 300s)
        """
        self.cache[key] = {
            'value': copy.deepcopy(value),
            'expires': time.time() + duration
        }

    def lookup(self, key):
        """Return dict with value or None if not present or expired.

        :param str key: Key for value

        Returns {'value': <value>} or None
        """
        res = None

        if key in self.cache:
            entry = self.cache[key]
            # check expiry
            if time.time() < entry['expires']:
                # return value
                res = {'value':  copy.deepcopy(entry['value'])}
            else:
                # remove expired value
                del self.cache[key]

        return res


class Cache():
    """Nested dict for values where each key will expire after some time."""

    def __init__(self):
        self.init()

    def init(self):
        self.cache = {}

    def identity_keys(self, identity):
        """Return [group, username] for identity.

        :param obj identity: User name or Identity dict
        """
        if identity is not None:
            if isinstance(identity, dict):
                return [
                    identity.get('group'),
                    identity.get('username')
                ]
            else:
                # identity is username
                return [None, identity]
        else:
            # keys for empty user
            return [None, '_public_']

    def cache_entry(self, service, identity, keys):
        # cache is a nested dict with the following levels:
        cache_keys = [service] + self.identity_keys(identity) + list(keys)
        # print("cache_entry %s" % cache_keys)
        cache = self.cache

        # read or initialize cache level by level
        for key in cache_keys[:-2]:
            entry = cache.get(key)
            if entry is None:
                entry = {}
                cache[key] = entry
            cache = entry

        # second last key points to ExpiringDict cache entry
        key = cache_keys[-2]
        entry = cache.get(key)
        if entry is None:
            entry = ExpiringDict()
            cache[key] = entry

        key = cache_keys[-1]
        return (entry, key)

    def read(self, service, identity, keys):
        cache, key = self.cache_entry(service, identity, keys)
        entry = cache.lookup(key)
        if entry:
            # print("Reading data from cache with key '%s'" % key)
            return entry['value']
        else:
            return None

    def write(self, service, identity, keys, data,
              cache_duration):
        cache, key = self.cache_entry(service, identity, keys)
        # print("Writing data into cache with key '%s'" % key)
        cache.set(key, data, cache_duration)



import os

from sqlalchemy import create_engine


class DatabaseEngine():
    """Helper for database connections using SQLAlchemy engines"""

    def __init__(self):
        """Constructor"""
        self.engines = {}

    def db_engine(self, conn_str):
        """Return engine.

        :param str conn_str: DB connection string for SQLAlchemy engine

        see http://docs.sqlalchemy.org/en/latest/core/engines.html#postgresql
        """
        engine = self.engines.get(conn_str)
        if not engine:
            engine = create_engine(
                conn_str, pool_pre_ping=True, echo=False)
            self.engines[conn_str] = engine
        return engine

    def db_engine_env(self, env_name, default=None):
        """Return engine configured in environment variable.

        :param str env_name: Environment variable
        :param str default: Default value if environment variable is not set
        """
        conn_str = os.environ.get(env_name, default)
        if conn_str is None:
            raise Exception(
                'db_engine_env: Environment variable %s not set' % env_name)
        return self.db_engine(conn_str)

    def geo_db(self):
        """Return engine for default GeoDB."""
        return self.db_engine_env('GEODB_URL',
                                  'postgresql:///?service=qwc_geodb')

    def config_db(self):
        """Return engine for default ConfigDB."""
        return self.db_engine_env('CONFIGDB_URL',
                                  'postgresql:///?service=qwc_configdb')
